/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Vehicle/route.h>
#include <Vehicle/vehicle.h>
#include <TopoFace/fgeomobjs.h>
#include <string.h>

Route::Route(const char* name, int npts, float* x, float* y, float* z) {
  _name = name ? strdup(name) : nil; 
  _xy = new FMultiLineObj(x, y, npts);
  if (!z) {
    z = new float[npts];
    for (int i=0; i<npts; i++) z[i] = 0.0;
  }
  float* d = new float[npts];
  for (int i=0; i<npts-1; i++) {
    double dx = x[i+1] - x[i];
    double dy = y[i+1] - y[i];
    double dz = z[i+1] - z[i];
    d[i] = sqrt(dx*dx + dy*dy + dz*dz);
  }
  _zd = new FMultiLineObj(z, d, npts);
  _npts = npts;
  _curdir = 0;
  _curx = _cury = _curz = 0.0;
  _curpt = 0;
  _vhcl = nil;
}

Route::~Route() {
  delete _name;
  delete _xy;
  delete _zd;
}

float* Route::x() const { return _xy ? _xy->x() : nil; }
float* Route::y() const { return _xy ? _xy->y() : nil; }
float* Route::z() const { return _zd ? _zd->x() : nil; }
float* Route::d() const { return _zd ? _zd->y() : nil; }

void Route::clear() {
  delete _xy;
  _xy = new FMultiLineObj(nil, nil, 0);
  delete _zd;
  _zd = new FMultiLineObj(nil, nil, 0);
  _npts = 0;
  _curdir = 0;
  _curx = _cury = _curz = 0.0;
  _curpt = 0;
  _vhcl = nil;
}

void Route::append(float xval, float yval, float zval) {
  if (_xy->size()<=npts()) {
      _xy->GrowActualBuf();
      _zd->GrowActualBuf();
  }
  x()[npts()] = xval;
  y()[npts()] = yval;
  z()[npts()] = zval;

  int prev = npts()-1;
  float dx = xval - x()[prev];
  float dy = yval - y()[prev];
  float dz = zval - z()[prev];

  d()[npts()] = sqrt(dx*dx + dy*dy + dz*dz);

  _npts++;
}

void Route::start(Vehicle* vehicle, boolean reverse) {
  _curdir = reverse ? -1 : 1;
  _curpt = reverse ? npts()-1 : 0;
  _curx = x()[_curpt];
  _cury = y()[_curpt];
  _curz = z()[_curpt];
  _vhcl = vehicle;
}

void Route::reset() {
  _curpt = _curdir==-1 ? npts()-1 : 0;
  _curx = x()[_curpt];
  _cury = y()[_curpt];
  _curz = z()[_curpt];
  _segdist = 0.0;
}

void Route::step(float meters) {
  if (meters<0) return;

  float dx = 0.0;
  float dy = 0.0;
  float dz = 0.0;
    
  while (!finished() && meters>0) {
    float x0 = x()[_curpt];
    float y0 = y()[_curpt];
    float z0 = z() ? z()[_curpt] : 0.0;
    
    float x1 = x()[_curpt+_curdir];
    float y1 = y()[_curpt+_curdir];
    float z1 = z() ? z()[_curpt+_curdir] : 0.0;
    
    float dist = d()[_curpt];
    float step = meters;
    boolean nextpt_flag = false;
    if (_segdist+meters>=dist) {
      step = dist-_segdist;
      nextpt_flag = true;
    } 
    _segdist += step;
    meters -= step;

    dx = x1-x0;
    dy = y1-y0;
    dz = z1-z0;
    
    double percent = _segdist/dist;
    
    _curx = x0 + dx * percent;
    _cury = y0 + dy * percent;
    _curz = z0 + dz * percent;

    if (nextpt_flag) {
      _segdist = 0.0;
      _curpt += _curdir;
      int endpt = _curdir>0 ? npts()-1 : 0;
      if (_curpt==endpt) {
	_curdir = 0;
	meters = 0.0;
      }
    }
  }
    
  _vhcl->turnto(dx, dy);
  _vhcl->jump(_curx, _cury, _curz);
}

boolean Route::finished() {
  return _curdir == 0;
}
