/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Time/obstime.h>
#include <Vehicle/vehicle.h>
#include <Vehicle/route.h>
#include <string.h>

Vehicle::Vehicle(const char* name, float x, float y, float z,
		 float dx, float dy, float dz) {
    _name = name ? strdup(name) : nil; 
    xloc() = x;
    yloc() = y;
    zloc() = z;
    xdir() = dx;
    ydir() = dy;
    zdir() = dz;
    _rte = nil;
    _tag = nil;
    _speed = 0;
    _clock = nil;
    _start = nil;
}

Vehicle::~Vehicle() {
    delete _name;
    delete _clock;
    delete _start;
}

void Vehicle::move(float dx, float dy) {
    xloc() += dx;
    yloc() += dy;
    growroute();
    notify();
}

void Vehicle::move(float dx, float dy, float dz) {
    xloc() += dx;
    yloc() += dy;
    zloc() += dz;
    growroute();
    notify();
}

void Vehicle::move(float* deltas, int ndeltas) {
    for (int i=0; i<min(3,ndeltas); i++) 
        _loc[i] += deltas[i];
    growroute();
    notify();
}

void Vehicle::jump(float x, float y) {
    xloc() = x;
    yloc() = y;
    growroute();
    notify();
}

void Vehicle::jump(float x, float y, float z) {
    xloc() = x;
    yloc() = y;
    zloc() = z;
    growroute();
    notify();
}

void Vehicle::jump(float* loc, int n) {
    for (int i=0; i<min(3,n); i++) 
        _loc[i] += loc[i];
    growroute();
    notify();
}

void Vehicle::turn(float dx, float dy) {
    xdir() += dx;
    ydir() += dy;
    notify();
}

void Vehicle::turn(float dx, float dy, float dz) {
    xdir() += dx;
    ydir() += dy;
    zdir() += dz;
    notify();
}

void Vehicle::turn(float* deltas, int ndeltas) {
    for (int i=0; i<min(3,ndeltas); i++) 
        _dir[i] += deltas[i];
    notify();
}

void Vehicle::turnto(float x, float y) {
    xdir() = x;
    ydir() = y;
    notify();
}

void Vehicle::turnto(float x, float y, float z) {
    xdir() = x;
    ydir() = y;
    zdir() = z;
    notify();
}

void Vehicle::turnto(float* dir, int n) {
    for (int i=0; i<min(3,n); i++) 
        _dir[i] += dir[i];
    notify();
}

void Vehicle::forward2d(float dist) {
    _loc[0] += _dir[0] * dist;
    _loc[1] += _dir[1] * dist;
    growroute();
    notify();
}

void Vehicle::backward2d(float dist) {
    _loc[0] -= _dir[0] * dist;
    _loc[1] -= _dir[1] * dist;
    growroute();
    notify();
}

void Vehicle::rotate2d(float ang) {
    vec2 vec(_dir[0], _dir[1]);
    mat3 mat = rotation2D(vec, ang);
    vec2 newvec = vec * mat;
    _dir[0] = newvec[0];
    _dir[1] = newvec[1];
    notify();
}

float Vehicle::heading2d() {
  if (xdir() < 0) {
      if (ydir() < 0) 
	  return 180.0 + atan(-xdir()/-ydir()) * RAD_TO_DEG;
      else if (ydir() == 0) 
	  return 270.0;
      else /* ydir() > 0 */
	   return 270 + atan(ydir()/-xdir()) * RAD_TO_DEG;
  } else if (xdir() == 0) {
      if (ydir() < 0) 
	  return 180.0;
      else /* ydir() >= 0 */
	   return 0.0;
  } else /* xdir() > 0 */ {
      if (ydir() < 0) 
	  return 90.0 + atan(-ydir()/xdir()) * RAD_TO_DEG;
      else if (ydir() == 0) 
	  return 90.0;
      else 
	  return atan(xdir()/ydir()) * RAD_TO_DEG;
  }
}

void Vehicle::startroute(Route* route) {
    _rte = route;
    growroute();
}

void Vehicle::stoproute() {
    _rte = nil;
}

void Vehicle::growroute() {
    if (_rte && !_start) 
        _rte->append(xloc(), yloc(), zloc());
}

void Vehicle::update(Observable* obs) {
  if (!_rte) return;
  if (!_start) return;
  ObservableTime* obstime = (ObservableTime*) obs;
#if 0
  cerr << "time update for vehicle " << name() << ":  " << obstime->time()->seconds() << "\n";
#endif

  /* before start time */
  if (*_start > *obstime->time()) {
    if (_clock) {
      delete _clock;
      _clock = nil;
      reset();
    }
    return;
  }

  /* after start time */
  boolean first = false;
  if (!_clock) {
    _clock = new Time(*_start);
    first = true;
  }

  /* forward moving time */
  if (*obstime->time() >= *_clock) {
    unsigned long deltat = obstime->time()->seconds() - _clock->seconds();
    *_clock += deltat;
    if (deltat || first ) _rte->step(_speed*deltat);
  }  

  /* backward moving time */
  else {
    unsigned long deltat = _clock->seconds() - obstime->time()->seconds();
    *_clock -= deltat;
    _rte->reset();
    if (_clock->seconds() != _start->seconds())
      _rte->step(_speed*(_clock->seconds()-_start->seconds()));
  }

  notify();
}

void Vehicle::simroute(Route* route, Time* start) {
  _rte = route;
  _rte->start(this); 
  _start = new Time(*start);
}

void Vehicle::reset(float x, float y, float z,
		    float dx, float dy, float dz) {
  xloc() = x;
  yloc() = y;
  zloc() = z;
  xdir() = dx;
  ydir() = dy;
  zdir() = dz;
  if (_rte)
    _rte->reset();
}
