/*
 * Copyright (c) 1997-1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef vhclfunc_h
#define vhclfunc_h

#include <ComTerp/comfunc.h>

class ComTerp;
class ObservableTime;
class RouteComp;
class Time;
class VehicleComp;

#include <OS/table.h>
declareTable(VehicleTable,int,void*)
declareTable(RouteTable,int,void*)

class VhclFunc : public ComFunc {
public:
    VhclFunc(ComTerp*);
    virtual void execute() = 0;

    static VehicleTable* vhcltable() { return _vhcltable; }
    static RouteTable* rtetable() { return _rtetable; }
    static VehicleComp* findvhcl(int symid);
    static RouteComp* findrte(int symid);

protected:
    static VehicleTable* _vhcltable;
    static RouteTable* _rtetable;
};

class CreateVhclFunc : public VhclFunc {
public:
    CreateVhclFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhclsym) -- create and assign vehicle to variable"; }
};

class CreateCompVhclFunc : public VhclFunc {
public:
    CreateCompVhclFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(compvhclsym vhclsym1[,vhclsym2[,...vhclsymn]]) -- create a composite vehicle"; }
};

class ConnectVhclFunc : public VhclFunc {
public:
    ConnectVhclFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl_or_rte hoststr portnum) -- display vehicle or route on viewer"; }
};

class DisconnectVhclFunc : public VhclFunc {
public:
    DisconnectVhclFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl_or_rte) -- not implemented yet"; }
};

class MoveVhclFunc : public VhclFunc {
public:
    MoveVhclFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl dx dy :abs) -- move vehicle in utm coords"; }

protected:
    int _abs_symid;
};

class VhclForwardFunc : public VhclFunc {
public:
    VhclForwardFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl dist) -- move vehicle forward in its current direction "; }
};

class VhclBackwardFunc : public VhclFunc {
public:
    VhclBackwardFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl dist) -- move vehicle backward in its current direction "; }
};

class VhclRotateFunc : public VhclFunc {
public:
    VhclRotateFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl angle) -- rotate vehicle direction by the given angle"; }
};

class VhclViewerFunc : public VhclFunc {
public:
    VhclViewerFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(hoststr portnum) -- set up default viewer for vehicle or route connections"; }

protected:
    static ComValue* _host;
    static ComValue* _port;
};

class CreateRouteFunc : public VhclFunc {
public:
    CreateRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(waypts :name rtesym) -- create and assign route to variable"; }

protected:
    int _name_symid;
};

class StartRouteFunc : public VhclFunc {
public:
    StartRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(rte vhcl) -- start capturing route of a vehicle"; }

protected:
    int _name_symid;
};

class StopRouteFunc : public VhclFunc {
public:
    StopRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(rte vhcl) -- stop capturing route of a vehicle "; }
};

class PlayRouteFunc : public VhclFunc {
public:
    PlayRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(rte vhcl :speed mps) -- set up to play route at meters/second"; }

protected:
    int _speed_symid;
};

class StepRouteFunc : public VhclFunc {
public:
    StepRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(rte meters) -- move current position of route forward by meters"; }
};

class VhclSimTimeFunc : public VhclFunc {
public:
    VhclSimTimeFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(secs :abs :reset :back) -- increment simulation time, or set it absolutely,\nor reset to default, or set it backward."; }

    static ObservableTime* simtime() { return _simtime; }

protected:
    int _abs_symid;
    int _sec_symid;
    int _back_symid;
    int _reset_symid;

    static ObservableTime* _simtime;
};

class VhclLocFunc : public VhclFunc {
public:
    VhclLocFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(vhcl) -- print vehicle position"; }
};

class SimRouteFunc : public VhclFunc {
public:
    SimRouteFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "%s(rte vhcl startsec :speed mps) -- set up route to start playing at simulation time"; }

protected:
    int _speed_symid;
};

#endif
